#ifndef __IKD_TOOLS_PREFECTURERENDERER_H__
#define __IKD_TOOLS_PREFECTURERENDERER_H__

// s{`Ǘ

#include "ClientRenderer.h"
#include "PrefectureFactory.h"


namespace Tools {

	class PrefectureRenderer : public ClientRenderer {
	public:
		PrefectureRenderer() : hWnd_(), bShowBorder_( true ), bShowField_( true ), bShowVertex_( false ), bShowInfo_( true ), centerCoord_( 0.0, 0.0 ) {
			brush_ = CreateSolidBrush( 0x00BAE0B9 );
			vtxBrush_ = CreateSolidBrush( 0x00ff0000 );
			gridPen_ = CreatePen( PS_DOT, 1, 0x00a0a0a0 );
		}
		virtual ~PrefectureRenderer() {
			DeleteObject( brush_ );
			DeleteObject( vtxBrush_ );
		}

		// s{|S쐬
		bool init( HWND hWnd ) {
			if ( factory_.create() == false )
				return false;
			prefNames_ = factory_.getPrefectureNameList();
			curPref_ = prefNames_[ 0 ];
			hWnd_ = hWnd;
			return true;
		}

		// s{쐬҂擾
		const PrefectureFactory *getFactory() const {
			return &factory_;
		}

		// `悷s{w
		void selectPref( const std::string &prefName ) {
			setPref( prefName, getScreenWidth(), getScreenHeight() );
		}

		// E\H
		void showBorder( bool isShow ) {
			bShowBorder_ = isShow;
			InvalidateRect( hWnd_, 0, FALSE );
		}

		// E\Ԃ擾
		bool getShowBorder() const {
			return bShowBorder_;
		}

		// tB[h\H
		void showField( bool isShow ) {
			bShowField_ = isShow;
			InvalidateRect( hWnd_, 0, FALSE );
		}

		// tB[h\Ԃ擾
		bool getShowField() const {
			return bShowField_;
		}

		// _\H
		void showVertex( bool isShow ) {
			bShowVertex_ = isShow;
			InvalidateRect( hWnd_, 0, FALSE );
		}

		// _\Ԃ擾
		bool getShowVertex() const {
			return bShowVertex_;
		}

		// \H
		void showInfo( bool isShow ) {
			bShowInfo_ = isShow;
			InvalidateRect( hWnd_, 0, FALSE );
		}

		// \Ԃ擾
		bool getShowInfo() const {
			return bShowInfo_;
		}


	protected:
		// `悷s{ݒ
		void setPref( const std::string &pref, int sw, int sh ) {
			partsPoints_.clear();
			curPref_ = pref;
			auto prefData = factory_.getPrefectureVertices( curPref_ );
			if ( prefData == 0 )
				return;	// [hIĂȂ

			centerCoord_ = prefData->center_;

			Double2 len = prefData->max_ - prefData->min_;
			double scale = ( len.y / len.x > (double)sh / sw ) ? sh / len.y: sw / len.x;
			scale *= 0.95;	// Ək߂

			for ( size_t i = 0; i < prefData->parts_.size(); ++i ) {
				auto part = prefData->parts_[ i ];
				// |S쐬
				std::vector< POINT > points;

				// ʑŜɎ܂TCYɒ
				// eWZ^[̑΋ 
				for ( size_t j = 0; j < part.points_.size(); ++j ) {
					Double2 dp = ( part.points_[ j ] - prefData->center_ ) * scale;
					POINT p = { sw / 2 + (LONG)dp.x, sh / 2 - (LONG)dp.y };
					points.push_back( p );
				}
				partsPoints_.push_back( points );
			}
			InvalidateRect( hWnd_, 0, FALSE );
		}

		// ʃTCY
		virtual void resize( HWND hwnd, HDC hdc, int w, int h ) {
			setPref( curPref_, w, h );
		}

		// `XV
		virtual void innerDraw( HWND hwnd, HDC hdc ) {
			HBRUSH brush = ( HBRUSH )::GetStockObject( NULL_BRUSH );
			if ( bShowField_ == true ) {
				brush = brush_;
			}
			auto preBrush = ::SelectObject( hdc, brush );
			HPEN pen = (HPEN)::GetStockObject( NULL_PEN );
			if ( bShowBorder_ == true ) {
				pen = ( HPEN )::GetStockObject( BLACK_PEN );
			}
			auto prePen = ::SelectObject( hdc, pen );
			for ( size_t i = 0; i < partsPoints_.size(); ++i ) {
				auto &points = partsPoints_[ i ];
				Polygon( hdc, &points[ 0 ], ( int )points.size() );
			}
			::SelectObject( hdc, preBrush );
			::SelectObject( hdc, prePen );

			// _`
			if ( bShowVertex_ == true ) {
				int r = 2;
				auto preBrush = ::SelectObject( hdc, vtxBrush_ );
				pen = ( HPEN )::GetStockObject( NULL_PEN );
				auto prePen = ::SelectObject( hdc, pen );
				for ( size_t i = 0; i < partsPoints_.size(); ++i ) {
					auto &points = partsPoints_[ i ];
					for ( size_t j = 0; j < points.size(); ++j ) {
						auto &p = points[ j ];
						::Ellipse( hdc, p.x - r, p.y - r, p.x + r, p.y + r );
					}
				}
				::SelectObject( hdc, preBrush );
				::SelectObject( hdc, prePen );
			}

			// CZX
			int w = getScreenWidth();
			int h = getScreenHeight();
			std::string licenseText( "yn@nn}{2.1ŃxN^[f[^ɍ쐬" );
			::TextOutA( hdc, w - ( int )licenseText.size() * 8 - 8, h - 20, licenseText.c_str(), ( int )licenseText.size() );

			// \
			if ( bShowInfo_ == true ) {
				SetBkMode( hdc, TRANSPARENT );
				std::string prefName( "Prefecture: " );
				prefName += curPref_;
				::TextOutA( hdc, 10, 10, prefName.c_str(), ( int )strlen( prefName.c_str() ) );

				char c[ 64 ];
				sprintf_s( c, 64, "%f", centerCoord_.y );
				::TextOutA( hdc, 10, h / 2 - 20, c, ( int )strlen( c ) );
				sprintf_s( c, 64, "%f", centerCoord_.x );
				::TextOutA( hdc, w / 2 + 4, 10, c, ( int )strlen( c ) );

				prePen = ::SelectObject( hdc, gridPen_ );
				::MoveToEx( hdc, 0, h / 2, 0 );
				::LineTo( hdc, w, h / 2 );
				::MoveToEx( hdc, w / 2, 0, 0 );
				::LineTo( hdc, w / 2, h );
				::SelectObject( hdc, prePen );
			}
		}

	private:
		PrefectureFactory factory_;
		std::vector< std::string > prefNames_;
		std::string curPref_;
		std::vector< std::vector< POINT > > partsPoints_;
		HWND hWnd_;
		HBRUSH brush_;
		HBRUSH vtxBrush_;
		HPEN gridPen_;
		bool bShowBorder_;
		bool bShowField_;
		bool bShowVertex_;
		bool bShowInfo_;
		Double2 centerCoord_;
	};
}

#endif
